/*
 * Copyright (c) 2011, Jeannette Bohg and Mårten Björkman 
 * ({bohg,celle}@csc.kth.se) 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  1.Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  2.Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.  
 *  3.The name of Jeannette Bohg or Mårten Björkman may not be used to 
 *    endorse or promote products derived from this software without 
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <ros/ros.h>

#include <fast_plane_detection/Plane.h>
#include <fast_plane_detection/PlaneInRegion.h>

#include <fast_plane_detector/utils.h>

#include <geometry_msgs/Vector3Stamped.h>
#include <visualization_msgs/Marker.h>

int main(int argc, char **argv)
{
  /** 
   * Init the node and give it a name. 
   */
  ros::init(argc, argv, "detect_plane_in_region");
  if (argc != 4)
    {
    ROS_INFO("usage: detect_plane_in_region X Y Z");
    return 1;
  }

  // get a ros handle
  ros::NodeHandle n;
  // create a client for the service add_two_ints
  ros::ServiceClient client 
    = n.serviceClient<fast_plane_detection::PlaneInRegion>("fast_plane_detection");
  // service class AddTwoInts is instantiated
  
  //! Publisher to Plane Parameters
  ros::Publisher plane_marker_pub_;
  // publish plane markers
  plane_marker_pub_ 
    = n.advertise<visualization_msgs::Marker>("region_plane_markers", 1); 

  fast_plane_detection::PlaneInRegion srv;
  geometry_msgs::Vector3Stamped point;
  point.header.frame_id = "narrow_stereo_optical_frame";
  point.header.stamp = ros::Time::now();
  point.vector.x = atoll(argv[1]);
  point.vector.y = atoll(argv[2]);
  point.vector.z = atoll(argv[3]);
  // request members are filled
  srv.request.point  = point;
  srv.request.width  = 10;
  srv.request.height = 50;

  int marker_id = 345;

  // service call that is blocking until call is actually made
  // is returning true if response is received
  if (client.call(srv))
  {
    ROS_INFO("Service call returned successfully");
    ROS_INFO("srv.response.plane.top_left %f %f %f", 
	     srv.response.plane.top_left.x, 
	     srv.response.plane.top_left.y,
	     srv.response.plane.top_left.z);
    // first send delete marker
    
    
    visualization_msgs::Marker delete_marker;
    delete_marker.header.stamp = ros::Time::now();
    delete_marker.header.frame_id = point.header.frame_id;
    delete_marker.id = marker_id;
    delete_marker.action = visualization_msgs::Marker::DELETE;
    delete_marker.ns = "tabletop_node";
    plane_marker_pub_.publish(delete_marker);
    

    // send current marker
    visualization_msgs::Marker planeMarker 
      = fast_plane_detection::getPlaneMarker( srv.response.plane.top_left, 
					      srv.response.plane.top_right,
					      srv.response.plane.bottom_left, 
					      srv.response.plane.bottom_right);
    planeMarker.header.frame_id = point.header.frame_id;
    planeMarker.header.stamp = ros::Time::now();
    planeMarker.pose = srv.response.plane.pose.pose;
    planeMarker.ns = "tabletop_node";
    planeMarker.id = marker_id; //arbitrary
    plane_marker_pub_.publish(planeMarker);
    

  }  else  {
    ROS_ERROR("Failed to call service fast_plane_detection");
    return 1;
  }
  
  return 0;
}
